@@ -6,7 +6,7 @@
  *                                                                         *
  * Copyright 2011 David Cleaver                                            *
  *                                                                         *
- * Copyright 2012, 2013, 2014, 2015 Case Van Horsen                        *
+ * Copyright 2012 - 2022 Case Van Horsen                                   *
  *                                                                         *
  * The original file is available at:                                      *
  *   <http://sourceforge.net/projects/mpzprp/files/>                       *
@@ -46,7 +46,7 @@ PyDoc_STRVAR(doc_mpz_is_fermat_prp,
 static PyObject *
 GMPY_mpz_is_fermat_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *a = NULL, *n = NULL;
+    MPZ_Object *a = NULL, *n = NULL;
     PyObject *result = NULL;
     mpz_t res, nm1;
 
@@ -55,14 +55,11 @@ GMPY_mpz_is_fermat_prp(PyObject *self, PyObject *args)
         return NULL;
     }
 
-    /* Allocate the temporary values early to allow a single cleanup section
-     * to be used. */
+    mpz_init(res);
+    mpz_init(nm1);
 
-    mpz_inoc(res);
-    mpz_inoc(nm1);
-
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
-    a = Pympz_From_Integer(PyTuple_GET_ITEM(args, 1));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
+    a = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 1), NULL);
     if (!a || !n) {
         TYPE_ERROR("is_fermat_prp() requires 2 integer arguments");
         goto cleanup;
@@ -74,8 +71,9 @@ GMPY_mpz_is_fermat_prp(PyObject *self, PyObject *args)
         goto cleanup;
     }
 
-    if (mpz_cmp_ui(n->z, 2) < 0) {
-        result = Py_False;
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_fermat_prp() requires 'n' be greater than 0");
         goto cleanup;
     }
 
@@ -98,7 +96,7 @@ GMPY_mpz_is_fermat_prp(PyObject *self, PyObject *args)
     /* Check gcd(a,n) */
     mpz_gcd(res, n->z, a->z);
     if (mpz_cmp_ui(res, 1) > 0) {
-        result = Py_False;
+        VALUE_ERROR("is_fermat_prp() requires gcd(n,a) == 1");
         goto cleanup;
     }
 
@@ -113,8 +111,8 @@ GMPY_mpz_is_fermat_prp(PyObject *self, PyObject *args)
 
   cleanup:
     Py_XINCREF(result);
-    mpz_cloc(res);
-    mpz_cloc(nm1);
+    mpz_clear(res);
+    mpz_clear(nm1);
     Py_XDECREF((PyObject*)a);
     Py_XDECREF((PyObject*)n);
     return result;
@@ -139,7 +137,7 @@ PyDoc_STRVAR(doc_mpz_is_euler_prp,
 static PyObject *
 GMPY_mpz_is_euler_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *a = NULL, *n = NULL;
+    MPZ_Object *a = NULL, *n = NULL;
     PyObject *result = NULL;
     mpz_t res, exp;
     int ret;
@@ -149,14 +147,11 @@ GMPY_mpz_is_euler_prp(PyObject *self, PyObject *args)
         return NULL;
     }
 
-    /* Take advantage of the cache of mpz_t objects maintained by GMPY2 to
-     * avoid memory allocations. */
+    mpz_init(res);
+    mpz_init(exp);
 
-    mpz_inoc(res);
-    mpz_inoc(exp);
-
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
-    a = Pympz_From_Integer(PyTuple_GET_ITEM(args, 1));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
+    a = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 1), NULL);
     if (!a || !n) {
         TYPE_ERROR("is_euler_prp() requires 2 integer arguments");
         goto cleanup;
@@ -168,7 +163,14 @@ GMPY_mpz_is_euler_prp(PyObject *self, PyObject *args)
         goto cleanup;
     }
 
-    if (mpz_cmp_ui(n->z, 2) < 0) {
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_euler_prp() requires 'n' be greater than 0");
+        goto cleanup;
+    }
+
+    /* Check for n == 1 */
+    if (mpz_cmp_ui(n->z, 1) == 0) {
         result = Py_False;
         goto cleanup;
     }
@@ -185,7 +187,7 @@ GMPY_mpz_is_euler_prp(PyObject *self, PyObject *args)
     /* Check gcd(a,b) */
     mpz_gcd(res, n->z, a->z);
     if (mpz_cmp_ui(res, 1) > 0) {
-        result = Py_False;
+        VALUE_ERROR("is_euler_prp() requires gcd(n,a) == 1");
         goto cleanup;
     }
 
@@ -210,8 +212,8 @@ GMPY_mpz_is_euler_prp(PyObject *self, PyObject *args)
 
   cleanup:
     Py_XINCREF(result);
-    mpz_cloc(res);
-    mpz_cloc(exp);
+    mpz_clear(res);
+    mpz_clear(exp);
     Py_XDECREF((PyObject*)a);
     Py_XDECREF((PyObject*)n);
     return result;
@@ -239,25 +241,22 @@ PyDoc_STRVAR(doc_mpz_is_strong_prp,
 static PyObject *
 GMPY_mpz_is_strong_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *a = NULL, *n = NULL;
+    MPZ_Object *a = NULL, *n = NULL;
     PyObject *result = NULL;
     mpz_t s, nm1, mpz_test;
-    mpir_ui r = 0;
+    mp_bitcnt_t r = 0;
 
     if (PyTuple_Size(args) != 2) {
         TYPE_ERROR("is_strong_prp() requires 2 integer arguments");
         return NULL;
     }
 
-    /* Take advantage of the cache of mpz_t objects maintained by GMPY2 to
-     * avoid memory allocations. */
+    mpz_init(s);
+    mpz_init(nm1);
+    mpz_init(mpz_test);
 
-    mpz_inoc(s);
-    mpz_inoc(nm1);
-    mpz_inoc(mpz_test);
-
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
-    a = Pympz_From_Integer(PyTuple_GET_ITEM(args, 1));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
+    a = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 1), NULL);
     if (!a || !n) {
         TYPE_ERROR("is_strong_prp() requires 2 integer arguments");
         goto cleanup;
@@ -269,7 +268,14 @@ GMPY_mpz_is_strong_prp(PyObject *self, PyObject *args)
         goto cleanup;
     }
 
-    if (mpz_cmp_ui(n->z, 2) < 0) {
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_strong_prp() requires 'n' be greater than 0");
+        goto cleanup;
+    }
+
+    /* Check for n == 1 */
+    if (mpz_cmp_ui(n->z, 1) == 0) {
         result = Py_False;
         goto cleanup;
     }
@@ -283,6 +289,13 @@ GMPY_mpz_is_strong_prp(PyObject *self, PyObject *args)
         goto cleanup;
     }
 
+    /* Check gcd(a,b) */
+    mpz_gcd(s, n->z, a->z);
+    if (mpz_cmp_ui(s, 1) > 0) {
+        VALUE_ERROR("is_strong_prp() requires gcd(n,a) == 1");
+        goto cleanup;
+    }
+
     mpz_set(nm1, n->z);
     mpz_sub_ui(nm1, nm1, 1);
 
@@ -290,6 +303,7 @@ GMPY_mpz_is_strong_prp(PyObject *self, PyObject *args)
     r = mpz_scan1(nm1, 0);
     mpz_fdiv_q_2exp(s, nm1, r);
 
+
     /* Check a^((2^t)*s) mod n for 0 <= t < r */
     mpz_powm(mpz_test, a->z, s, n->z);
     if ((mpz_cmp_ui(mpz_test, 1) == 0) || (mpz_cmp(mpz_test, nm1) == 0)) {
@@ -311,9 +325,9 @@ GMPY_mpz_is_strong_prp(PyObject *self, PyObject *args)
     result = Py_False;
   cleanup:
     Py_XINCREF(result);
-    mpz_cloc(s);
-    mpz_cloc(nm1);
-    mpz_cloc(mpz_test);
+    mpz_clear(s);
+    mpz_clear(nm1);
+    mpz_clear(mpz_test);
     Py_XDECREF((PyObject*)a);
     Py_XDECREF((PyObject*)n);
     return result;
@@ -339,32 +353,29 @@ PyDoc_STRVAR(doc_mpz_is_fibonacci_prp,
 static PyObject *
 GMPY_mpz_is_fibonacci_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *n = NULL, *p = NULL, *q = NULL;
+    MPZ_Object *n = NULL, *p = NULL, *q = NULL;
     PyObject *result = NULL;
     mpz_t pmodn, zP;
     /* used for calculating the Lucas V sequence */
     mpz_t vl, vh, ql, qh, tmp;
-    size_t s = 0, j = 0;
+    mp_bitcnt_t s = 0, j = 0;
 
     if (PyTuple_Size(args) != 3) {
         TYPE_ERROR("is_fibonacci_prp() requires 3 integer arguments");
         return NULL;
     }
 
-    /* Take advantage of the cache of mpz_t objects maintained by GMPY2 to
-     * avoid memory allocations. */
+    mpz_init(pmodn);
+    mpz_init(zP);
+    mpz_init(vl);
+    mpz_init(vh);
+    mpz_init(ql);
+    mpz_init(qh);
+    mpz_init(tmp);
 
-    mpz_inoc(pmodn);
-    mpz_inoc(zP);
-    mpz_inoc(vl);
-    mpz_inoc(vh);
-    mpz_inoc(ql);
-    mpz_inoc(qh);
-    mpz_inoc(tmp);
-
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
-    p = Pympz_From_Integer(PyTuple_GET_ITEM(args, 1));
-    q = Pympz_From_Integer(PyTuple_GET_ITEM(args, 2));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
+    p = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 1), NULL);
+    q = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 2), NULL);
     if (!n || !p || !q) {
         TYPE_ERROR("is_fibonacci_prp() requires 3 integer arguments");
         goto cleanup;
@@ -387,7 +398,14 @@ GMPY_mpz_is_fibonacci_prp(PyObject *self, PyObject *args)
         goto cleanup;
     }
 
-    if (mpz_cmp_ui(n->z, 2) < 0) {
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_fibonacci_prp() requires 'n' be greater than 0");
+        goto cleanup;
+    }
+
+    /* Check for n == 1 */
+    if (mpz_cmp_ui(n->z, 1) == 0) {
         result = Py_False;
         goto cleanup;
     }
@@ -485,13 +503,13 @@ GMPY_mpz_is_fibonacci_prp(PyObject *self, PyObject *args)
 
   cleanup:
     Py_XINCREF(result);
-    mpz_cloc(pmodn);
-    mpz_cloc(zP);
-    mpz_cloc(vl);
-    mpz_cloc(vh);
-    mpz_cloc(ql);
-    mpz_cloc(qh);
-    mpz_cloc(tmp);
+    mpz_clear(pmodn);
+    mpz_clear(zP);
+    mpz_clear(vl);
+    mpz_clear(vh);
+    mpz_clear(ql);
+    mpz_clear(qh);
+    mpz_clear(tmp);
     Py_XDECREF((PyObject*)p);
     Py_XDECREF((PyObject*)q);
     Py_XDECREF((PyObject*)n);
@@ -518,12 +536,12 @@ PyDoc_STRVAR(doc_mpz_is_lucas_prp,
 static PyObject *
 GMPY_mpz_is_lucas_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *n= NULL, *p = NULL, *q = NULL;
+    MPZ_Object *n = NULL, *p = NULL, *q = NULL;
     PyObject *result = NULL;
     mpz_t zD, res, index;
     /* used for calculating the Lucas U sequence */
     mpz_t uh, vl, vh, ql, qh, tmp;
-    size_t s = 0, j = 0;
+    mp_bitcnt_t s = 0, j = 0;
     int ret;
 
     if (PyTuple_Size(args) != 3) {
@@ -531,22 +549,19 @@ GMPY_mpz_is_lucas_prp(PyObject *self, PyObject *args)
         return NULL;
     }
 
-    /* Take advantage of the cache of mpz_t objects maintained by GMPY2 to
-     * avoid memory allocations. */
+    mpz_init(zD);
+    mpz_init(res);
+    mpz_init(index);
+    mpz_init(uh);
+    mpz_init(vl);
+    mpz_init(vh);
+    mpz_init(ql);
+    mpz_init(qh);
+    mpz_init(tmp);
 
-    mpz_inoc(zD);
-    mpz_inoc(res);
-    mpz_inoc(index);
-    mpz_inoc(uh);
-    mpz_inoc(vl);
-    mpz_inoc(vh);
-    mpz_inoc(ql);
-    mpz_inoc(qh);
-    mpz_inoc(tmp);
-
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
-    p = Pympz_From_Integer(PyTuple_GET_ITEM(args, 1));
-    q = Pympz_From_Integer(PyTuple_GET_ITEM(args, 2));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
+    p = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 1), NULL);
+    q = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 2), NULL);
     if (!n || !p || !q) {
         TYPE_ERROR("is_lucas_prp() requires 3 integer arguments");
         goto cleanup;
@@ -561,7 +576,14 @@ GMPY_mpz_is_lucas_prp(PyObject *self, PyObject *args)
         goto cleanup;
     }
 
-    if (mpz_cmp_ui(n->z, 2) < 0) {
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_lucas_prp() requires 'n' be greater than 0");
+        goto cleanup;
+    }
+
+    /* Check for n == 1 */
+    if (mpz_cmp_ui(n->z, 1) == 0) {
         result = Py_False;
         goto cleanup;
     }
@@ -580,7 +602,7 @@ GMPY_mpz_is_lucas_prp(PyObject *self, PyObject *args)
     mpz_mul_ui(res, res, 2);
     mpz_gcd(res, res, n->z);
     if ((mpz_cmp(res, n->z) != 0) && (mpz_cmp_ui(res, 1) > 0)) {
-        result = Py_False;
+        VALUE_ERROR("is_lucas_prp() requires gcd(n,2*q*D) == 1");
         goto cleanup;
     }
 
@@ -728,37 +750,33 @@ PyDoc_STRVAR(doc_mpz_is_stronglucas_prp,
 static PyObject *
 GMPY_mpz_is_stronglucas_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *n = NULL, *p = NULL, *q = NULL;
+    MPZ_Object *n = NULL, *p= NULL, *q = NULL;
     PyObject *result = NULL;
     mpz_t zD, s, nmj, res;
     /* these are needed for the LucasU and LucasV part of this function */
     mpz_t uh, vl, vh, ql, qh, tmp;
-    mpir_ui r = 0;
+    mp_bitcnt_t r = 0, j = 0;
     int ret = 0;
-    mpir_ui j = 0;
 
     if (PyTuple_Size(args) != 3) {
         TYPE_ERROR("is_strong_lucas_prp() requires 3 integer arguments");
         return NULL;
     }
 
-    /* Take advantage of the cache of mpz_t objects maintained by GMPY2 to
-     * avoid memory allocations. */
+    mpz_init(zD);
+    mpz_init(s);
+    mpz_init(nmj);
+    mpz_init(res);
+    mpz_init(uh);
+    mpz_init(vl);
+    mpz_init(vh);
+    mpz_init(ql);
+    mpz_init(qh);
+    mpz_init(tmp);
 
-    mpz_inoc(zD);
-    mpz_inoc(s);
-    mpz_inoc(nmj);
-    mpz_inoc(res);
-    mpz_inoc(uh);
-    mpz_inoc(vl);
-    mpz_inoc(vh);
-    mpz_inoc(ql);
-    mpz_inoc(qh);
-    mpz_inoc(tmp);
-
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
-    p = Pympz_From_Integer(PyTuple_GET_ITEM(args, 1));
-    q = Pympz_From_Integer(PyTuple_GET_ITEM(args, 2));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
+    p = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 1), NULL);
+    q = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 2), NULL);
     if (!n || !p || !q) {
         TYPE_ERROR("is_strong_lucas_prp() requires 3 integer arguments");
         goto cleanup;
@@ -773,7 +791,14 @@ GMPY_mpz_is_stronglucas_prp(PyObject *self, PyObject *args)
         goto cleanup;
     }
 
-    if (mpz_cmp_ui(n->z, 2) < 0) {
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_strong_lucas_prp() requires 'n' be greater than 0");
+        goto cleanup;
+    }
+
+    /* Check for n == 1 */
+    if (mpz_cmp_ui(n->z, 1) == 0) {
         result = Py_False;
         goto cleanup;
     }
@@ -792,7 +817,7 @@ GMPY_mpz_is_stronglucas_prp(PyObject *self, PyObject *args)
     mpz_mul_ui(res, res, 2);
     mpz_gcd(res, res, n->z);
     if ((mpz_cmp(res, n->z) != 0) && (mpz_cmp_ui(res, 1) > 0)) {
-        result = Py_False;
+        VALUE_ERROR("is_strong_lucas_prp() requires gcd(n,2*q*D) == 1");
         goto cleanup;
     }
 
@@ -941,47 +966,42 @@ PyDoc_STRVAR(doc_mpz_is_extrastronglucas_prp,
 "    gcd(n, 2*D) == 1\n"
 "    n = s*(2**r) + Jacobi(D,n), s odd\n"
 "Then an extra strong Lucas probable prime requires:\n"
-"    (lucasu(p,1,s) == 0 (mod n)\n"
-"     and\n"
-"     lucasv(p,1,s) == +/-2 (mod n))\n"
+"    lucasu(p,1,s) == 0 (mod n)\n"
+"    or\n"
+"    lucasv(p,1,s) == +/-2 (mod n)\n"
 "    or\n"
 "    lucasv(p,1,s*(2**t)) == 0 (mod n) for some t, 0 <= t < r");
 
 static PyObject *
 GMPY_mpz_is_extrastronglucas_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *n = NULL, *p = NULL;
+    MPZ_Object *n = NULL, *p = NULL;
     PyObject *result = NULL;
     mpz_t zD, s, nmj, nm2, res;
     /* these are needed for the LucasU and LucasV part of this function */
     mpz_t uh, vl, vh, ql, qh, tmp;
-    mpir_ui r = 0;
-    mpir_si q = 1;
+    mp_bitcnt_t r = 0, j = 0;
     int ret = 0;
-    size_t j = 0;
 
     if (PyTuple_Size(args) != 2) {
         TYPE_ERROR("is_extra_strong_lucas_prp() requires 2 integer arguments");
         return NULL;
     }
 
-    /* Take advantage of the cache of mpz_t objects maintained by GMPY2 to
-     * avoid memory allocations. */
+    mpz_init(zD);
+    mpz_init(s);
+    mpz_init(nmj);
+    mpz_init(nm2);
+    mpz_init(res);
+    mpz_init(uh);
+    mpz_init(vl);
+    mpz_init(vh);
+    mpz_init(ql);
+    mpz_init(qh);
+    mpz_init(tmp);
 
-    mpz_inoc(zD);
-    mpz_inoc(s);
-    mpz_inoc(nmj);
-    mpz_inoc(nm2);
-    mpz_inoc(res);
-    mpz_inoc(uh);
-    mpz_inoc(vl);
-    mpz_inoc(vh);
-    mpz_inoc(ql);
-    mpz_inoc(qh);
-    mpz_inoc(tmp);
-
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
-    p = Pympz_From_Integer(PyTuple_GET_ITEM(args, 1));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
+    p = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 1), NULL);
     if (!n || !p) {
         TYPE_ERROR("is_extra_strong_lucas_prp() requires 2 integer arguments");
         goto cleanup;
@@ -991,11 +1011,18 @@ GMPY_mpz_is_extrastronglucas_prp(PyObject *self, PyObject *args)
     mpz_mul(zD, p->z, p->z);
     mpz_sub_ui(zD, zD, 4);
     if (mpz_sgn(zD) == 0) {
-        VALUE_ERROR("is_extra_strong_lucas_prp() requires p*p-4 != 0");
+        VALUE_ERROR("invalid value for p in is_extra_strong_lucas_prp()");
         goto cleanup;
     }
 
-    if (mpz_cmp_ui(n->z, 2) < 0) {
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_extra_strong_lucas_prp() requires 'n' be greater than 0");
+        goto cleanup;
+    }
+
+    /* Check for n == 1 */
+    if (mpz_cmp_ui(n->z, 1) == 0) {
         result = Py_False;
         goto cleanup;
     }
@@ -1013,7 +1040,7 @@ GMPY_mpz_is_extrastronglucas_prp(PyObject *self, PyObject *args)
     mpz_mul_ui(res, zD, 2);
     mpz_gcd(res, res, n->z);
     if ((mpz_cmp(res, n->z) != 0) && (mpz_cmp_ui(res, 1) > 0)) {
-        result = Py_False;
+        VALUE_ERROR("is_extra_strong_lucas_prp() requires gcd(n,2*D) == 1");
         goto cleanup;
     }
 
@@ -1028,10 +1055,10 @@ GMPY_mpz_is_extrastronglucas_prp(PyObject *self, PyObject *args)
     r = mpz_scan1(nmj, 0);
     mpz_fdiv_q_2exp(s, nmj, r);
 
-    /* (-2 mod n) is needed for one of the tests. */
-    mpz_sub_ui(nm2, n->z, 2);
+    mpz_set(nm2, n->z);
+    mpz_sub_ui(nm2, nm2, 2);
 
-    /* make sure that either U_s == 0 mod n and V_s == +/-2 mod n, or */
+    /* make sure that either U_s == 0 mod n or V_s == +/-2 mod n, or */
     /* V_((2^t)*s) == 0 mod n for some t with 0 <= t < r-1           */
     mpz_set_si(uh, 1);
     mpz_set_si(vl, 2);
@@ -1046,7 +1073,7 @@ GMPY_mpz_is_extrastronglucas_prp(PyObject *self, PyObject *args)
         mpz_mod(ql, ql, n->z);
         if (mpz_tstbit(s,j) == 1) {
             /* qh = ql*q */
-            mpz_mul_si(qh, ql, q);
+            mpz_set(qh, ql);
 
             /* uh = uh*vh (mod n) */
             mpz_mul(uh, uh, vh);
@@ -1108,8 +1135,8 @@ GMPY_mpz_is_extrastronglucas_prp(PyObject *self, PyObject *args)
     mpz_mod(vl, vl, n->z);
 
     /* uh contains LucasU_s and vl contains LucasV_s */
-    if ((mpz_cmp_ui(uh, 0) == 0) &&
-       ((mpz_cmp(vl, nm2) == 0) || (mpz_cmp_si(vl, 2) == 0))) {
+    if ((mpz_cmp_ui(uh, 0) == 0) || (mpz_cmp_ui(vl, 0) == 0) ||
+        (mpz_cmp(vl, nm2) == 0) || (mpz_cmp_si(vl, 2) == 0)) {
         result = Py_True;
         goto cleanup;
     }
@@ -1169,7 +1196,7 @@ PyDoc_STRVAR(doc_mpz_is_selfridge_prp,
 static PyObject *
 GMPY_mpz_is_selfridge_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *n = NULL;
+    MPZ_Object *n = NULL;
     PyObject *result = NULL, *temp = NULL;
     long d = 5, p = 1, q = 0, max_d = 1000000;
     int jacobi = 0;
@@ -1180,18 +1207,22 @@ GMPY_mpz_is_selfridge_prp(PyObject *self, PyObject *args)
         return NULL;
     }
 
-    /* Take advantage of the cache of mpz_t objects maintained by GMPY2 to
-     * avoid memory allocations. */
+    mpz_init(zD);
 
-    mpz_inoc(zD);
-
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
     if (!n) {
         TYPE_ERROR("is_selfridge_prp() requires 1 integer argument");
         goto cleanup;
     }
 
-    if (mpz_cmp_ui(n->z, 2) < 0) {
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_selfridge_prp() requires 'n' be greater than 0");
+        goto cleanup;
+    }
+
+    /* Check for n == 1 */
+    if (mpz_cmp_ui(n->z, 1) == 0) {
         result = Py_False;
         goto cleanup;
     }
@@ -1205,7 +1236,6 @@ GMPY_mpz_is_selfridge_prp(PyObject *self, PyObject *args)
         goto cleanup;
     }
 
-
     mpz_set_ui(zD, d);
 
     while (1) {
@@ -1267,8 +1297,8 @@ GMPY_mpz_is_selfridge_prp(PyObject *self, PyObject *args)
   cleanup:
     Py_XINCREF(result);
   return_result:
-    mpz_cloc(zD);
-    Py_XDECREF((PyObject*)n);
+    mpz_clear(zD);
+    Py_DECREF((PyObject*)n);
     return result;
 }
 
@@ -1276,7 +1306,7 @@ GMPY_mpz_is_selfridge_prp(PyObject *self, PyObject *args)
  * mpz_strongselfridge_prp:
  * A "strong Lucas-Selfridge probable prime" n is a "strong Lucas probable prime" using Selfridge parameters of:
  * Find the first element D in the sequence {5, -7, 9, -11, 13, ...} such that Jacobi(D,n) = -1
- * Then use P=1 and Q=(1-D)/4 in the strong Lucase probable prime test.
+ * Then use P=1 and Q=(1-D)/4 in the strong Lucas probable prime test.
  * Make sure n is not a perfect square, otherwise the search for D will only stop when D=n.
  * **********************************************************************************************************/
 
@@ -1291,7 +1321,7 @@ PyDoc_STRVAR(doc_mpz_is_strongselfridge_prp,
 static PyObject *
 GMPY_mpz_is_strongselfridge_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *n = NULL;
+    MPZ_Object *n = NULL;
     PyObject *result = NULL, *temp = NULL;
     long d = 5, p = 1, q = 0, max_d = 1000000;
     int jacobi = 0;
@@ -1302,18 +1332,22 @@ GMPY_mpz_is_strongselfridge_prp(PyObject *self, PyObject *args)
         return NULL;
     }
 
-    /* Take advantage of the cache of mpz_t objects maintained by GMPY2 to
-     * avoid memory allocations. */
+    mpz_init(zD);
 
-    mpz_inoc(zD);
-
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
     if (!n) {
         TYPE_ERROR("is_strong_selfridge_prp() requires 1 integer argument");
         goto cleanup;
     }
 
-    if (mpz_cmp_ui(n->z, 2) < 0) {
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_strong_selfridge_prp() requires 'n' be greater than 0");
+        goto cleanup;
+    }
+
+    /* Check for n == 1 */
+    if (mpz_cmp_ui(n->z, 1) == 0) {
         result = Py_False;
         goto cleanup;
     }
@@ -1327,6 +1361,7 @@ GMPY_mpz_is_strongselfridge_prp(PyObject *self, PyObject *args)
         goto cleanup;
     }
 
+
     mpz_set_ui(zD, d);
 
     while (1) {
@@ -1388,8 +1423,8 @@ GMPY_mpz_is_strongselfridge_prp(PyObject *self, PyObject *args)
   cleanup:
     Py_XINCREF(result);
   return_result:
-    mpz_cloc(zD);
-    Py_XDECREF((PyObject*)n);
+    mpz_clear(zD);
+    Py_DECREF((PyObject*)n);
     return result;
 }
 
@@ -1409,7 +1444,7 @@ PyDoc_STRVAR(doc_mpz_is_bpsw_prp,
 static PyObject *
 GMPY_mpz_is_bpsw_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *n = NULL;
+    MPZ_Object *n = NULL;
     PyObject *result = NULL, *temp = NULL;
 
     if (PyTuple_Size(args) != 1) {
@@ -1417,12 +1452,33 @@ GMPY_mpz_is_bpsw_prp(PyObject *self, PyObject *args)
         return NULL;
     }
 
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
     if (!n) {
         TYPE_ERROR("is_bpsw_prp() requires 1 integer argument");
         goto cleanup;
     }
 
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_bpsw_prp() requires 'n' be greater than 0");
+        goto cleanup;
+    }
+
+    /* Check for n == 1 */
+    if (mpz_cmp_ui(n->z, 1) == 0) {
+        result = Py_False;
+        goto cleanup;
+    }
+
+    /* Handle n even. */
+    if (mpz_divisible_ui_p(n->z, 2)) {
+        if (mpz_cmp_ui(n->z, 2) == 0)
+            result = Py_True;
+        else
+            result = Py_False;
+        goto cleanup;
+    }
+
     /* "O" is used to increment the reference to n so deleting temp won't
      * delete n.
      */
@@ -1446,7 +1502,7 @@ GMPY_mpz_is_bpsw_prp(PyObject *self, PyObject *args)
   cleanup:
     Py_XINCREF(result);
   return_result:
-    Py_XDECREF((PyObject*)n);
+    Py_DECREF((PyObject*)n);
     return result;
 }
 
@@ -1462,25 +1518,46 @@ PyDoc_STRVAR(doc_mpz_is_strongbpsw_prp,
 "is_strong_bpsw_prp(n) -> boolean\n\n"
 "Return True if n is a strong Baillie-Pomerance-Selfridge-Wagstaff\n"
 "probable prime. A strong BPSW probable prime passes the is_strong_prp()\n"
-"test with base 2 and the is_strong_selfridge_prp() test.\n");
+"test with base and the is_strong_selfridge_prp() test.\n");
 
 static PyObject *
 GMPY_mpz_is_strongbpsw_prp(PyObject *self, PyObject *args)
 {
-    PympzObject *n;
-    PyObject *result = 0, *temp = 0;
+    MPZ_Object *n = NULL;
+    PyObject *result = NULL, *temp = NULL;
 
     if (PyTuple_Size(args) != 1) {
         TYPE_ERROR("is_strong_bpsw_prp() requires 1 integer argument");
         return NULL;
     }
 
-    n = Pympz_From_Integer(PyTuple_GET_ITEM(args, 0));
+    n = GMPy_MPZ_From_Integer(PyTuple_GET_ITEM(args, 0), NULL);
     if (!n) {
         TYPE_ERROR("is_strong_bpsw_prp() requires 1 integer argument");
         goto cleanup;
     }
 
+    /* Require n > 0. */
+    if (mpz_sgn(n->z) <= 0) {
+        VALUE_ERROR("is_strong_bpsw_prp() requires 'n' be greater than 0");
+        goto cleanup;
+    }
+
+    /* Check for n == 1 */
+    if (mpz_cmp_ui(n->z, 1) == 0) {
+        result = Py_False;
+        goto cleanup;
+    }
+
+    /* Handle n even. */
+    if (mpz_divisible_ui_p(n->z, 2)) {
+        if (mpz_cmp_ui(n->z, 2) == 0)
+            result = Py_True;
+        else
+            result = Py_False;
+        goto cleanup;
+    }
+
     /* "O" is used to increment the reference to n so deleting temp won't
      * delete n.
      */
@@ -1497,14 +1574,14 @@ GMPY_mpz_is_strongbpsw_prp(PyObject *self, PyObject *args)
     temp = Py_BuildValue("(O)", n);
     if (!temp)
         goto cleanup;
-    result = GMPY_mpz_is_strongselfridge_prp(NULL, temp);
+    result = GMPY_mpz_is_selfridge_prp(NULL, temp);
     Py_DECREF(temp);
     goto return_result;
 
   cleanup:
     Py_XINCREF(result);
   return_result:
-    Py_XDECREF((PyObject*)n);
+    Py_DECREF((PyObject*)n);
     return result;
 }
 
